import {navigation} from '../../../../src/shared/utils/navigation.utils';
import * as utils from '../../../utils';

const url = 'https://earthquake.usgs.gov/ws/nshmp/data/gmm/magnitude';

describe('Ground Motion vs. Magnitude', () => {
  beforeEach(() => {
    cy.viewport('macbook-16');
    cy.intercept(url).as('usage');
    cy.visit('/gmm/magnitude');
    cy.wait('@usage');
  });

  describe('Check nshmp-template', () => {
    utils.hasNshmpTemplate(navigation());
  });

  describe('Application', () => {
    utils.hasControlPanel();
    utils.hasPlotContent();
    utils.toggleControlPanel();
    utils.togglePlotPanel();
    utils.toggleSettingPanel();
    utils.changePlotSettings('.mean-plot', '.mean-settings');
    utils.changePlotSettings('.sigma-plot', '.sigma-settings');
    utils.hasGmmMenu();

    it('Can set GMM menus', () => {
      setGmmMenus();
    });

    it('Can set event parameters', () => {
      setEventParameters();
    });

    it('Can set geometry parameters', () => {
      utils.setSourceGeometryParameters();
    });

    it('Can set site parameters', () => {
      utils.setSiteParameters();
    });

    it('Can set path parameters', () => {
      setPathParameters();
    });

    it('Can reset control panel inputs', () => {
      utils.canResetControlPanel(() => setControlPanel());
    });

    it('Calls service and plots response', () => {
      const buttons: utils.ControlPanelButtons = {
        hasResetButton: false,
        hasServiceCallButton: true,
        hasSubmitButton: true,
      };
      utils.controlPanelButtonsNotEnabled(buttons);

      setControlPanel();
      utils.submitFormCheckPlots({
        firstPlotClass: '.mean-plot',
        secondPlotClass: '.sigma-plot',
      });

      utils.controlPanelButtonsEnabled(buttons);
    });

    describe('Application Tabs', () => {
      it('Has plot tab', () => {
        cy.get('plotly-plot');

        utils.hasApplicationTab('.medians-tab');

        cy.get('plotly-plot').should('not.exist');

        utils.hasApplicationTab('.plots-tab');

        cy.get('plotly-plot');
      });

      it('Has medians tab', () => {
        cy.get('plotly-plot');

        utils.exportDataTableNotExist();
        utils.dataTableNotExist();

        setControlPanel();

        utils.submitFormCheckPlots({
          firstPlotClass: '.mean-plot',
          secondPlotClass: '.sigma-plot',
        });

        utils.hasApplicationTab('.medians-tab');

        cy.get('plotly-plot').should('not.exist');

        utils.hasExportDataTable('gmm-magnitude-means.csv');
        utils.hasDataTable();
      });

      it('Has sigmas tab', () => {
        cy.get('plotly-plot');

        utils.exportDataTableNotExist();
        utils.dataTableNotExist();

        setControlPanel();

        utils.submitFormCheckPlots({
          firstPlotClass: '.mean-plot',
          secondPlotClass: '.sigma-plot',
        });

        utils.hasApplicationTab('.sigmas-tab');

        cy.get('plotly-plot').should('not.exist');

        utils.hasExportDataTable('gmm-magnitude-sigmas.csv');
        utils.hasDataTable();
      });
    });
  });
});

function setControlPanel() {
  setGmmMenus();
  setEventParameters();
  utils.setSourceGeometryParameters();
  setPathParameters();
  utils.setSiteParameters();
}

function setGmmMenus() {
  cy.get('nshmp-template-control-panel')
    .find('form')
    .should('be.visible')
    .as('form');

  cy.get('@form').find('.multi-select-parameter-menu').click();
  utils.selectMatOption('gmm');

  cy.get('@form').find('.imt-select').find('mat-select').click();
  utils.selectMatOption('default');

  cy.get('@form')
    .find('gmm-lib-gmm-menu')
    .find('select')
    .select([0, 1, 2, 3, 4]);

  cy.get('@form')
    .find('.imt-select')
    .find('mat-select')
    .click()
    .get('.cdk-overlay-container')
    .find('mat-option')
    .should('not.have.value', 'default');
  utils.selectMatOption('PGA');
}

function setPathParameters() {
  cy.get('nshmp-template-control-panel')
    .find('.distance-input')
    .find('input')
    .should('be.visible')
    .clear()
    .type('20.0')
    .should('have.value', '20.0');
}

function setEventParameters() {
  cy.get('nshmp-template-control-panel')
    .find('.mmin-input')
    .find('input')
    .should('be.visible')
    .clear()
    .type('4.0')
    .should('have.value', '4.0');

  cy.get('nshmp-template-control-panel')
    .find('.mmax-input')
    .find('input')
    .should('be.visible')
    .clear()
    .type('7.5')
    .should('have.value', '7.5');

  cy.get('nshmp-template-control-panel')
    .find('.mstep-input')
    .find('input')
    .should('be.visible')
    .clear()
    .type('0.1')
    .should('have.value', '0.1');
}
