import 'cypress';

import {NavigationList} from '@ghsc/nshmp-template';

export interface CheckPlot {
  plotClass: string;
  intercept?: boolean;
}

export interface CheckPlots {
  firstPlotClass: string;
  secondPlotClass: string;

  intercept?: boolean;
}

export function canResetControlPanel(setControlPanel: () => unknown) {
  setControlPanel();

  cy.get('nshmp-lib-control-panel-buttons')
    .find('.reset-button')
    .find('button')
    .click();
}

export function clickSettingsPanel() {
  cy.get('nshmp-template-app-controls').find('.settings-toggle').click();
}

export function hasControlPanel() {
  it('Has control panel', () => {
    cy.get('nshmp-template-control-panel').should('be.visible');
  });
}

export function hasNshmpTemplate(navList: NavigationList[]) {
  it('Has nshmp-template', () => {
    cy.get('nshmp-template').should('be.visible');
  });

  it('Has nshmp-template header', () => {
    cy.get('nshmp-template-header').should('be.visible');
  });

  it('Has nshmp-template header logo', () => {
    cy.get('div.nshmp-template-header--logo')
      .should('be.visible')
      .find('img')
      .should('be.visible')
      .should('have.attr', 'src', 'assets/usgs-logo.svg');
  });

  it('Return to dashboard when logo clicked', () => {
    cy.get('div.nshmp-template-header--logo')
      .should('be.visible')
      .find('a')
      .should('be.visible')
      .click()
      .url()
      .should('eq', Cypress.config().baseUrl);
  });

  describe('Navigation Dropdown', () => {
    it('Has navigation dropdown', () => {
      cy.get('nshmp-template-navigation')
        .find('button')
        .click()
        .get('.mat-menu-panel')
        .should('be.visible')
        .find('button')
        .should('have.length.above', 0);
    });

    navList.forEach(nav => {
      if (nav.subHeader) {
        it(`Has "${nav.subHeader}" header`, () => {
          cy.get('nshmp-template-navigation')
            .find('button')
            .click()
            .get('.mat-menu-panel')
            .contains(nav.subHeader);
        });
      }
      nav.navigation.forEach(app => {
        const routerLink = app.routerLink.startsWith('/')
          ? app.routerLink.substring(1)
          : app.routerLink;

        it(`Has "${app.display}" listed in navigation menu`, () => {
          if (app.routerLink.includes('/aws/')) {
            cy.intercept(
              'GET',
              'https://earthquake.usgs.gov/ws/nshmp/hazard-runs/auth',
              {
                fixture: 'aws-auth.json',
              }
            );
          }

          cy.get('nshmp-template-navigation').find('button').click();

          cy.get('.mat-menu-panel')
            .then(menu => {
              const subMenu = menu.find(`button:contains('${nav.subHeader}')`);
              if (subMenu.length) {
                cy.get('.mat-menu-panel')
                  .find(`button:contains('${nav.subHeader}')`)
                  .trigger('mouseenter');
              }
            })
            .get('.mat-menu-panel')
            .contains('button', app.display)
            .click()
            .url()
            .should('contain', `${Cypress.config().baseUrl}${routerLink}`);
        });
      });
    });
  });
}

export function hasPlotContent() {
  it('Has plot content', () => {
    cy.get('nshmp-template-plot-content').should('exist');
  });
}

export function submitFormCheckPlot(options: CheckPlot) {
  options.intercept =
    options.intercept === undefined ? true : options.intercept;

  cy.get('nshmp-template-plot-content')
    .find(options.plotClass)
    .then(originalPlot => {
      if (options.intercept) {
        cy.intercept('**').as('service-call');
      }

      cy.get('nshmp-lib-control-panel-buttons')
        .find('.submit-button')
        .find('button')
        .click();

      if (options.intercept) {
        cy.wait('@service-call');
      }

      cy.get('nshmp-template-plot-content')
        .find(options.plotClass)
        .should(updatedPlot => {
          expect(originalPlot).to.not.equal(updatedPlot);
        });
    });
}

export function submitFormCheckPlots(options: CheckPlots) {
  options.intercept =
    options.intercept === undefined ? true : options.intercept;

  cy.get('nshmp-template-plot-content')
    .find(options.firstPlotClass)
    .then(originalFirstPlot => {
      cy.get('nshmp-template-plot-content')
        .find(options.secondPlotClass)
        .then(originalSecondPlot => {
          if (options.intercept) {
            cy.intercept('**').as('service-call');
          }

          cy.get('nshmp-lib-control-panel-buttons')
            .find('.submit-button')
            .find('button')
            .click();

          if (options.intercept) {
            cy.wait('@service-call', {
              timeout: 30000,
            });
          }

          cy.get('nshmp-template-plot-content')
            .find(options.firstPlotClass)
            .should(updatedFirstPlot => {
              expect(originalFirstPlot).to.not.equal(updatedFirstPlot);
            });
          cy.get('nshmp-template-plot-content')
            .find(options.secondPlotClass)
            .should(updatedSecondPlot => {
              expect(originalSecondPlot).to.not.equal(updatedSecondPlot);
            });
        });
    });
}

export function toggleControlPanel() {
  it('Toggles control panel', () => {
    cy.get('nshmp-template-app-controls')
      .find('.control-panel-toggle')
      .should('have.class', 'selected')
      .should('be.enabled')
      .as('control-panel-button');

    cy.get('nshmp-template-control-panel').should('be.visible');
    cy.get('@control-panel-button').click();
    cy.get('nshmp-template-control-panel').should('not.exist');
  });
}

export function togglePlotPanel() {
  it('Toggles plot content', () => {
    cy.get('nshmp-template-app-controls')
      .find('.plot-toggle')
      .should('have.class', 'selected')
      .should('be.enabled')
      .as('plot-button');

    cy.get('nshmp-template-plot-content').should('exist');
    cy.get('@plot-button').click();
    cy.get('nshmp-template-plot-content').should('not.exist');
  });
}

export function toggleSettingPanel() {
  it('Toggles setting panel', () => {
    cy.get('nshmp-template-app-controls')
      .find('.settings-toggle')
      .should('not.have.class', 'selected')
      .should('be.enabled')
      .as('settings-button');

    cy.get('nshmp-template-settings').should('not.exist');
    cy.get('@settings-button').click();
    cy.get('nshmp-template-settings').find('form').should('be.visible');
  });
}
