import {AsyncPipe, DatePipe, NgClass, UpperCasePipe} from '@angular/common';
import {
  AfterViewInit,
  Component,
  effect,
  ElementRef,
  OnDestroy,
  ViewChild,
} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatButtonToggleGroup} from '@angular/material/button-toggle';
import {MatIcon} from '@angular/material/icon';
import {MatProgressSpinner} from '@angular/material/progress-spinner';
import {
  MatSlideToggle,
  MatSlideToggleChange,
} from '@angular/material/slide-toggle';
import {MatSnackBar} from '@angular/material/snack-bar';
import {MatToolbar, MatToolbarRow} from '@angular/material/toolbar';
import {MatTooltip} from '@angular/material/tooltip';
import {JobStatus} from '@ghsc/nshmp-lib-ng/aws';
import {NshmpTemplateService} from '@ghsc/nshmp-template';

import {GetLogsResponseData} from '../../models/get-logs.model';
import {LogPipe} from '../../pipes/log.pipe';
import {AppService} from '../../services/app.service';
import {SnackBarComponent} from '../snack-bar/snack-bar.component';

/**
 * Main application content showing logs.
 */
@Component({
  imports: [
    MatToolbar,
    MatToolbarRow,
    MatProgressSpinner,
    MatSlideToggle,
    MatTooltip,
    NgClass,
    UpperCasePipe,
    DatePipe,
    LogPipe,
    ReactiveFormsModule,
    MatIcon,
    MatButtonToggleGroup,
    AsyncPipe,
  ],
  selector: 'app-content',
  standalone: true,
  styleUrl: './content.component.scss',
  templateUrl: './content.component.html',
})
export class ContentComponent implements OnDestroy, AfterViewInit {
  /** Job status enum */
  jobStatus = JobStatus;

  /** Element reference to logs element */
  @ViewChild('logs')
  logsEl: ElementRef<HTMLDivElement>;

  /** The log service response */
  serviceResponse = this.service.serviceResponse;
  /** Control form state */
  form = this.service.formGroup;
  /** Whether service is loading */
  loading = this.service.serviceLoading;

  constructor(
    public templateServie: NshmpTemplateService,
    private service: AppService,
    private snackBar: MatSnackBar
  ) {
    effect(
      () => {
        if (this.service.serviceResponse()) {
          this.scrollBottom();
          this.checkLogs(this.service.serviceResponse().response);
        }
      },
      {
        allowSignalWrites: true,
      }
    );
  }

  ngOnDestroy(): void {
    this.snackBar.dismiss();
  }

  ngAfterViewInit(): void {
    this.scrollBottom();
  }

  /**
   * Handle slide toogle change for application auto refreshe
   */
  onRefreshChange(event: MatSlideToggleChange): void {
    if (event.checked) {
      this.service.callService(false);
    }
  }

  /**
   * Scroll to bottom of page for logs.
   */
  scrollBottom(): void {
    /* Wait for log to display and scroll */
    setTimeout(() => {
      if (this.logsEl) {
        this.logsEl.nativeElement.scroll({
          top: this.logsEl.nativeElement.scrollHeight,
        });
      }
    }, 0);
  }

  /**
   * Scroll to top of logs.
   */
  scrollTop() {
    this.logsEl.nativeElement.scroll({top: 0});
  }

  /**
   * Check log response for errors or results.
   *
   * @param response The logs service response
   */
  private checkLogs(response: GetLogsResponseData): void {
    if (response.jobStatus.toString() === JobStatus.FAILED.toString()) {
      alert('nshmp-haz failed, check email for more information');
    }

    if (response.results) {
      // Call to get last logs
      const lastLog = [...response.logs].pop();
      if (!lastLog.message.includes('Finished')) {
        this.service.callService(false);
      }

      this.snackBar.openFromComponent(SnackBarComponent, {
        data: response.results,
      });
    }
  }
}
