import {computed, Injectable, Signal, signal} from '@angular/core';
import {AbstractControl, FormBuilder, Validators} from '@angular/forms';
import {ActivatedRoute, Router} from '@angular/router';
import {
  AwsService,
  JobHistoryRequestData,
  JobStatus,
  TerminateJobRequestData,
} from '@ghsc/nshmp-lib-ng/aws';
import {NshmpService, SpinnerService} from '@ghsc/nshmp-lib-ng/nshmp';
import {DynamoDBItem} from '@ghsc/nshmp-utils-ts/libs/aws/run-nshmp-haz';
import {Response} from '@ghsc/nshmp-utils-ts/libs/nshmp-ws-utils';
import {EC2} from 'aws-sdk';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {AppServiceModel} from 'projects/nshmp-apps/src/shared/models/app-service.model';
import {devApps} from 'projects/nshmp-apps/src/shared/utils/applications.utils';
import {catchError} from 'rxjs';

import {ControlForm} from '../models/control-form.model';
import {
  GetLogsRequestData,
  GetLogsResponseData,
} from '../models/get-logs.model';
import {AppState} from '../models/state.model';

/**
 * check-haz-jobs application service.
 */
@Injectable({
  providedIn: 'root',
})
export class AppService implements AppServiceModel<AppState, ControlForm> {
  private baseUrl = environment.webServices.aws.url;
  private hazardJobsService = environment.webServices.aws.services.hazardJobs;
  private hazardLogsService = environment.webServices.aws.services.hazardLogs;
  private terminateService = environment.webServices.aws.services.terminateJob;

  /** Reactive form group */
  readonly formGroup = this.formBuilder.group<ControlForm>(
    this.defaultFormValues()
  );
  /** Application signal state */
  readonly state = signal<AppState>(this.initialState());

  constructor(
    private formBuilder: FormBuilder,
    private spinnerService: SpinnerService,
    private nshmpService: NshmpService,
    private awsService: AwsService,
    private route: ActivatedRoute,
    private router: Router
  ) {
    this.addValidators();

    this.formGroup.valueChanges.subscribe(() => this.updateUrl());
  }

  /**
   * Returns the job info service response
   */
  get jobInfoResponse(): Signal<Response<JobHistoryRequestData, DynamoDBItem>> {
    return computed(() => this.state().jobHistoryResponse);
  }

  /**
   * Returns whether the log service is loading in background.
   */
  get serviceLoading(): Signal<boolean> {
    return computed(() => this.state().serviceLoading);
  }

  /**
   * Returns the log service response.
   */
  get serviceResponse(): Signal<
    Response<GetLogsRequestData, GetLogsResponseData>
  > {
    return computed(() => this.state().serviceResponse);
  }

  /**
   * Returns the terminate job service response.
   */
  get terminateJobResponse(): Signal<
    Response<TerminateJobRequestData, EC2.InstanceStateChangeList>
  > {
    return computed(() => this.state().terminateJobResponse);
  }

  /**
   * Add validators to form controls.
   */
  addValidators(): void {
    const required = (control: AbstractControl) => Validators.required(control);
    this.formGroup.controls.id.addValidators([
      required,
      Validators.pattern(/^[a-zA-Z0-9]*$/),
    ]);
    this.formGroup.controls.id.updateValueAndValidity();
  }

  /**
   * Call AWS job history service.
   */
  callJobInfoService(): void {
    this.awsService
      .callJobHistoryService<DynamoDBItem>(
        `${this.baseUrl}/${this.hazardJobsService}`,
        this.formGroup.controls.id
      )
      .subscribe(jobHistoryResponse => {
        this.updateState({jobHistoryResponse});

        this.awsService.openTerminateDialog(jobHistoryResponse, () =>
          this.callTerminateJobService()
        );
      });
  }

  /**
   * Call AWS log service.
   *
   * @param showSpinner Whether to show the spinner dialog
   */
  callService(showSpinner = true): void {
    const {id} = this.formGroup.getRawValue();

    if (showSpinner) {
      this.spinnerService.show(`Getting logs for job ${id}`);
    }

    const url = `${this.baseUrl}${this.hazardLogsService}/${id}`;
    this.updateState({
      serviceLoading: true,
    });

    this.nshmpService
      .callService$<Response<GetLogsRequestData, GetLogsResponseData>>(url)
      .pipe(catchError((error: Error) => this.nshmpService.throwError$(error)))
      .subscribe(serviceResponse =>
        this.handleServiceResponse(serviceResponse, showSpinner)
      );
  }

  defaultFormValues(): ControlForm {
    return {
      autoRefresh: true,
      id: '',
    };
  }

  /**
   * Initialize application.
   */
  init(): void {
    const query = this.route.snapshot.queryParams as ControlForm;
    const formValues: ControlForm = {
      ...this.defaultFormValues(),
      ...query,
    };

    this.formGroup.patchValue(formValues);

    if (this.formGroup.valid) {
      this.nshmpService.selectPlotControl();
      this.callService();
    }
  }

  initialState(): AppState {
    return {
      jobHistoryResponse: null,
      serviceLoading: false,
      serviceResponse: null,
      terminateJobResponse: null,
    };
  }

  /**
   * Update state.
   *
   * @param state Partial state to update
   */
  updateState(state: Partial<AppState>): void {
    this.state.set({
      ...this.state(),
      ...state,
    });
  }

  /**
   * Call the AWS terminate job service.
   */
  private callTerminateJobService(): void {
    this.awsService
      .callTerminateJobService(
        `${this.baseUrl}/${this.terminateService}`,
        this.formGroup.controls.id
      )
      .subscribe(terminateJobResponse =>
        this.updateState({terminateJobResponse})
      );
  }

  /**
   * Handle service response.
   *
   * @param serviceResponse The service response
   * @param showSpinner Whether to show spinner
   */
  private handleServiceResponse(
    serviceResponse: Response<GetLogsRequestData, GetLogsResponseData>,
    showSpinner: boolean
  ): void {
    const {autoRefresh} = this.formGroup.getRawValue();
    const status = serviceResponse.response.jobStatus;

    if (
      (status === JobStatus.RUNNING || status === JobStatus.CREATED) &&
      autoRefresh
    ) {
      setTimeout(() => this.callService(false), 3000);
    }

    if (showSpinner) {
      this.spinnerService.remove();
    }
    this.spinnerService.remove();

    this.updateState({
      serviceLoading: false,
      serviceResponse,
    });
  }

  /**
   * Update URL based on form values.
   */
  private updateUrl(): void {
    const query: Partial<ControlForm> = {
      id: this.formGroup.getRawValue().id,
    };

    this.router
      .navigate([devApps().aws.checkHazJobs.routerLink], {
        queryParams: {
          ...query,
        },
      })
      .catch((error: Error) => this.nshmpService.throwError$(error));
  }
}
