import {computed, Injectable, Signal, signal} from '@angular/core';
import {FormBuilder, Validators} from '@angular/forms';
import {ActivatedRoute, Router} from '@angular/router';
import {AwsService} from '@ghsc/nshmp-lib-ng/aws';
import {NshmpService, nshmpUtils} from '@ghsc/nshmp-lib-ng/nshmp';
import {DynamoDBItem} from '@ghsc/nshmp-utils-ts/libs/aws/run-nshmp-haz';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {AppServiceModel} from 'projects/nshmp-apps/src/shared/models/app-service.model';
import {devApps} from 'projects/nshmp-apps/src/shared/utils/applications.utils';

import {ControlForm} from '../models/control-form.model';
import {AppState} from '../models/state.model';

/**
 * Entrypoint for accessing store.
 */
@Injectable({
  providedIn: 'root',
})
export class AppService implements AppServiceModel<AppState, ControlForm> {
  private baseUrl = environment.webServices.aws.url;
  private jobsService = environment.webServices.aws.services.hazardJobs;

  /** Control form group */
  readonly formGroup = this.formBuilder.group<ControlForm>(
    this.defaultFormValues()
  );
  /** Application state signal */
  readonly state = signal<AppState>(this.initialState());

  constructor(
    private formBuilder: FormBuilder,
    private nshmpService: NshmpService,
    private awsService: AwsService,
    private route: ActivatedRoute,
    private router: Router
  ) {
    this.addValidators();
    this.formGroup.valueChanges.subscribe(() => this.updateUrl());
  }

  /**
   * Returns the job history filtered.
   */
  get jobHistory(): Signal<DynamoDBItem[]> {
    return computed(() => this.state().filteredHistory);
  }

  /**
   * Add form control validators.
   *
   * @override
   */
  addValidators(): void {
    this.formGroup.controls.id.addValidators(control =>
      Validators.required(control)
    );
    this.formGroup.controls.id.updateValueAndValidity();
  }

  /**
   * Dispatch action to call job info service.
   *
   * @override
   */
  callService(): void {
    this.awsService
      .callJobHistoryService<
        DynamoDBItem[]
      >(`${this.baseUrl}/${this.jobsService}`, this.formGroup.controls.id)
      .subscribe(jobHistoryResponse => {
        if (!Array.isArray(jobHistoryResponse.response)) {
          jobHistoryResponse = {
            ...jobHistoryResponse,
            response: [jobHistoryResponse.response],
          };
        }

        this.updateState({jobHistoryResponse});
        this.filterHistory();
      });
  }

  /**
   * Default form values
   *
   * @override
   */
  defaultFormValues(): ControlForm {
    return {
      filter: nshmpUtils.selectPlaceHolder().value,
      id: '',
    };
  }

  /**
   * Filter job history.
   */
  filterHistory(): void {
    const jobHistoryResponse = this.state().jobHistoryResponse;
    const filter = this.formGroup.getRawValue().filter;

    if (
      jobHistoryResponse === null ||
      filter === nshmpUtils.selectPlaceHolder().value
    ) {
      this.updateState({
        filteredHistory: jobHistoryResponse?.response ?? [],
      });
    } else {
      const filteredHistory = jobHistoryResponse.response.filter(
        jobInfo => jobInfo.job.jobStatus.toString() === filter
      );

      this.updateState({filteredHistory});
    }
  }

  /**
   * Initialize application.
   *
   * @override
   */
  init(): void {
    const query = this.route.snapshot.queryParams as ControlForm;
    const formValues: ControlForm = {
      ...this.defaultFormValues(),
      ...query,
    };

    this.formGroup.patchValue(formValues);

    if (this.formGroup.valid) {
      this.nshmpService.selectPlotControl();
      this.callService();
    }
  }

  /**
   * The initial application state.
   *
   * @override
   */
  initialState(): AppState {
    return {
      filteredHistory: [],
      jobHistoryResponse: null,
    };
  }

  /**
   * Update state
   *
   * @param state Partial state to update
   * @override
   */
  updateState(state: Partial<AppState>): void {
    this.state.set({
      ...this.state(),
      ...state,
    });
  }

  /**
   * Update URL based on form values.
   */
  private updateUrl(): void {
    const query: Partial<ControlForm> = {
      id: this.formGroup.getRawValue().id,
    };

    this.router
      .navigate([devApps().aws.hazJobHistory.routerLink], {
        queryParams: {
          ...query,
        },
      })
      .catch((error: Error) => this.nshmpService.throwError$(error));
  }
}
