import {NgClass} from '@angular/common';
import {Component} from '@angular/core';
import {MatDivider} from '@angular/material/divider';
import {
  MatAccordion,
  MatExpansionPanel,
  MatExpansionPanelHeader,
  MatExpansionPanelTitle,
} from '@angular/material/expansion';
import {
  MatCell,
  MatCellDef,
  MatColumnDef,
  MatHeaderCell,
  MatHeaderCellDef,
  MatHeaderRow,
  MatHeaderRowDef,
  MatRow,
  MatRowDef,
  MatTable,
} from '@angular/material/table';

import {commonInstances} from '../../models/instance-type.model';
import {mapRegions} from '../../models/map-regions.model';
import {NshmpHazClass} from '../../models/nshmp-haz-class.model';
import {AppService} from '../../services/app.service';

interface Parameter {
  name: string;
  type: string;

  indent?: true;
}

/**
 * Submit hazard job HTTP POST configuration parameters.
 *
 * Shows format of the POST JSON parameters to run hazard jobs.
 */
@Component({
  imports: [
    MatAccordion,
    MatExpansionPanel,
    MatExpansionPanelHeader,
    MatExpansionPanelTitle,
    MatTable,
    MatColumnDef,
    MatHeaderCellDef,
    MatHeaderCell,
    MatCellDef,
    MatCell,
    NgClass,
    MatHeaderRowDef,
    MatHeaderRow,
    MatRowDef,
    MatRow,
    MatDivider,
  ],
  selector: 'app-config',
  standalone: true,
  styleUrl: './config.component.scss',
  templateUrl: './config.component.html',
})
export class ConfigComponent {
  /** Common EC2 isntance types */
  commonInstances = commonInstances();
  /** Default values */
  defaultFormValues = this.service.defaultFormValues();

  /** Example configuration */
  exampleConfig = this.service.exampleConfig();

  exampleConfigArray = this.service.exampleConfigCalcArray();

  /** Calculation configuration example in JSON */
  calcConfigJson = `\n${JSON.stringify(this.exampleConfig.calcConfig, null, 2)}`;

  /** Calculation configuration example in JSON */
  calcConfigArrayJson = `\n${JSON.stringify(this.exampleConfigArray.calcConfig, null, 2)}`;

  /** Map region bounds values */
  mapRegions = mapRegions()
    .map(keyValue => keyValue.value)
    .filter(value => value !== null);

  /** nshmp-haz Java classes to run */
  NshmpHazClass = NshmpHazClass;

  /** Encoded URL to example site file */
  siteFileEncoded =
    'https://code.usgs.gov/api/v4/projects/ghsc%2Fnshmp%2Fnshmp-haz/' +
    'repository/files/etc%2Fexamples%2F3-sites-file%2Fsites.geojson/raw?ref=main';

  /** Angular table columns  */
  columns = ['name', 'type'];

  /** The calulation configuration parameters for POST call */
  calcConfig: Parameter[] = [
    {
      name: 'calcConfig',
      type: 'calcConfig | calcConfig[]',
    },
  ];

  /** Cloud configuration parameters for POST call */
  cloudConfig: Parameter[] = [
    {
      name: 'cloudConfig',
      type: '',
    },
    {
      indent: true,
      name: '.instanceType',
      type: 'string',
    },
  ];

  /** GMT hazard map configuration parameters for POST call */
  mapConfig: Parameter[] = [
    {
      name: 'mapConfig',
      type: '',
    },
    {
      indent: true,
      name: '.region',
      type: 'string',
    },
    {
      indent: true,
      name: '.title',
      type: 'string',
    },
    {
      indent: true,
      name: '.vs30',
      type: 'string',
    },
  ];

  /** NSHMP configuration parameters for POST call */
  nshmpConfig: Parameter[] = [
    {
      name: 'nshmpConfig',
      type: '',
    },
    {
      indent: true,
      name: '.className',
      type: 'string',
    },
    {
      indent: true,
      name: '.email',
      type: 'string',
    },
    {
      indent: true,
      name: '.modelGitUrl',
      type: 'string',
    },
    {
      indent: true,
      name: '.modelSha',
      type: 'string',
    },
    {
      indent: true,
      name: '.nshmpLibGitUrl',
      type: 'string',
    },
    {
      indent: true,
      name: '.nshmpLibSha',
      type: 'string',
    },
    {
      indent: true,
      name: '.siteFileUrl',
      type: 'string',
    },
    {
      indent: true,
      name: '.sourceCodeGitUrl',
      type: 'string',
    },
    {
      indent: true,
      name: '.sourceCodeSha',
      type: 'string',
    },
    {
      indent: true,
      name: '.modelPath',
      type: 'string',
    },
  ];

  /** All POST parameters */
  parameters: Parameter[] = [
    ...this.cloudConfig,
    ...this.nshmpConfig,
    ...this.mapConfig,
    ...this.calcConfig,
  ];

  constructor(private service: AppService) {}

  /**
   * Scroll to parameter info when clicked in table.
   *
   * @param configParameter The parameter to scroll to
   */
  onParameterClick(configParameter: string): void {
    const parameter = configParameter.split('.').pop() ?? configParameter;
    document.querySelector(`.${parameter}`)?.scrollIntoView();
  }
}
