import {AsyncPipe} from '@angular/common';
import {
  Component,
  effect,
  ElementRef,
  Input,
  OnDestroy,
  OnInit,
  ViewChild,
} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatButton, MatIconAnchor} from '@angular/material/button';
import {MatOption} from '@angular/material/core';
import {MatDialog} from '@angular/material/dialog';
import {MatDivider} from '@angular/material/divider';
import {
  MatAccordion,
  MatExpansionPanel,
  MatExpansionPanelDescription,
  MatExpansionPanelHeader,
  MatExpansionPanelTitle,
} from '@angular/material/expansion';
import {
  MatError,
  MatFormField,
  MatHint,
  MatLabel,
  MatPrefix,
  MatSuffix,
} from '@angular/material/form-field';
import {MatIcon} from '@angular/material/icon';
import {MatInput} from '@angular/material/input';
import {MatSelect} from '@angular/material/select';
import {MatSlideToggle} from '@angular/material/slide-toggle';
import {MatTab} from '@angular/material/tabs';
import {MatTooltip} from '@angular/material/tooltip';
import {NshmpTemplateFormFieldsComponent} from '@ghsc/nshmp-template';
import {CalcConfig} from '@ghsc/nshmp-utils-ts/libs/aws/run-nshmp-haz';
import {Subscription} from 'rxjs';
import * as YAML from 'yaml';

import {FileType} from '../../models/file-type.model';
import {commonInstances} from '../../models/instance-type.model';
import {mapRegions} from '../../models/map-regions.model';
import {NshmpHazClass} from '../../models/nshmp-haz-class.model';
import {AppService} from '../../services/app.service';
import {JobSubmittedComponent} from '../job-submitted/job-submitted.component';

/**
 * Form fields to run hazard on AWS.
 */
@Component({
  imports: [
    MatAccordion,
    MatExpansionPanel,
    MatExpansionPanelHeader,
    MatExpansionPanelTitle,
    MatDivider,
    MatFormField,
    MatLabel,
    MatInput,
    MatPrefix,
    MatIconAnchor,
    MatSuffix,
    MatTooltip,
    MatHint,
    MatError,
    MatSelect,
    MatOption,
    MatSlideToggle,
    MatExpansionPanelDescription,
    MatButton,
    AsyncPipe,
    NshmpTemplateFormFieldsComponent,
    ReactiveFormsModule,
    MatIcon,
  ],
  selector: 'app-form',
  standalone: true,
  styleUrl: './form.component.scss',
  templateUrl: './form.component.html',
})
export class FormComponent implements OnInit, OnDestroy {
  /** The angular material config tab  */
  @Input()
  configTab: MatTab;

  /** The angular material json example tab */
  @Input()
  jsonTab: MatTab;

  /** THe angular material yaml example tab */
  @Input()
  yamlTab: MatTab;

  /** The element reference for the calculation configuration input element */
  @ViewChild('calcConfigEl')
  calcConfigEl: ElementRef<HTMLInputElement>;

  /** The element reference for the configuration input element */
  @ViewChild('configEl')
  configEl: ElementRef<HTMLInputElement>;

  /** The element reference to the upload site file input */
  @ViewChild('uploadSiteFileEl')
  uploadSiteEl: ElementRef<HTMLInputElement>;

  /** Common EC2 instance types */
  commonInstances = commonInstances();
  /** Map region for GMT bounds */
  mapRegions = mapRegions();
  /** Form bounds */
  formBounds = this.service.formBounds();
  /** nshmp-haz Java classes to run */
  NshmpHazClass = NshmpHazClass;
  /** List of the Java classes */
  nshmpHazClasses = Object.values(NshmpHazClass);
  /** Whitespace reqex */
  whitespace = '/^\\S+$/';
  /** Whether the reset button is disabled */
  resetDisabled = false;

  /** Calculation configuration state */
  calcConfig = this.service.calcConfig;
  /** Control form state */
  form = this.service.formGroup;

  private overrideSubscription = new Subscription();

  constructor(
    public service: AppService,
    private matDialog: MatDialog
  ) {
    effect(() => {
      const responses = this.service.state().serviceResponses;

      if (responses && responses.length > 0) {
        this.matDialog.open(JobSubmittedComponent);
      }
    });
  }

  ngOnInit(): void {
    const {nshmpConfig} = this.form.controls;

    this.overrideSubscription =
      this.form.controls.overrideNshmpLib.valueChanges.subscribe(
        overrideNshmpLib => {
          if (overrideNshmpLib) {
            nshmpConfig.controls.nshmpLibGitUrl.enable();
            nshmpConfig.controls.nshmpLibSha.enable();
          } else {
            this.form.controls.nshmpConfig.controls.nshmpLibGitUrl.patchValue(
              null
            );
            nshmpConfig.controls.nshmpLibGitUrl.disable();
            nshmpConfig.controls.nshmpLibSha.patchValue(null);
            nshmpConfig.controls.nshmpLibSha.disable();
          }
        }
      );
  }

  ngOnDestroy(): void {
    this.overrideSubscription.unsubscribe();
  }

  /**
   * Return the calculation configuration as JSON string.
   *
   * @param calcConfig The calculation configuration
   * @returns
   */
  calcConfigJson(calcConfig: CalcConfig): string {
    return `\n${YAML.stringify(calcConfig, {indent: 2})}`;
  }

  /**
   * Dispatch action to import calculation configuration file.
   */
  importCalcConfigFile() {
    this.service.importCalcConfigFile(this.calcConfigEl.nativeElement.files);
  }

  /**
   * Dispatch aciton to import configuration file.
   */
  importConfigFile() {
    this.service.importConfigFile(this.configEl.nativeElement.files);
    this.configEl.nativeElement.value = '';
  }

  /**
   * Dispatch action to upload site file.
   */
  importSiteFile() {
    this.service.uploadFiles(
      this.uploadSiteEl.nativeElement.files,
      FileType.SITE
    );
    this.uploadSiteEl.nativeElement.value = '';
  }
}
