import {computed, Injectable, Signal, signal} from '@angular/core';
import {FormBuilder, Validators} from '@angular/forms';
import {ActivatedRoute, Router} from '@angular/router';
import {
  AwsService,
  IdForm,
  JobHistoryRequestData,
  TerminateJobRequestData,
} from '@ghsc/nshmp-lib-ng/aws';
import {NshmpService} from '@ghsc/nshmp-lib-ng/nshmp';
import {DynamoDBItem} from '@ghsc/nshmp-utils-ts/libs/aws/run-nshmp-haz';
import {Response} from '@ghsc/nshmp-utils-ts/libs/nshmp-ws-utils';
import {EC2} from 'aws-sdk';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {AppServiceModel} from 'projects/nshmp-apps/src/shared/models/app-service.model';
import {devApps} from 'projects/nshmp-apps/src/shared/utils/applications.utils';

import {AppState} from '../models/state.model';

@Injectable({
  providedIn: 'root',
})
export class AppService implements AppServiceModel<AppState, IdForm> {
  private baseUrl = environment.webServices.aws.url;
  private jobsService = environment.webServices.aws.services.hazardJobs;
  private terminateService = environment.webServices.aws.services.terminateJob;

  readonly formGroup = this.formBuilder.group<IdForm>(this.defaultFormValues());
  readonly state = signal<AppState>(this.initialState());

  constructor(
    private formBuilder: FormBuilder,
    private nshmpService: NshmpService,
    private awsService: AwsService,
    private route: ActivatedRoute,
    private router: Router
  ) {
    this.addValidators();
    this.formGroup.valueChanges.subscribe(() => this.updateUrl());
  }

  /**
   * Returns the job info service response observable.
   */
  get jobInfoResponse(): Signal<Response<JobHistoryRequestData, DynamoDBItem>> {
    return computed(() => this.state().jobHistoryResponse);
  }

  /**
   * Returns the terminate job service response observable.
   */
  get terminateJobResponse(): Signal<
    Response<TerminateJobRequestData, EC2.InstanceStateChangeList>
  > {
    return computed(() => this.state().terminateJobResponse);
  }

  /**
   * Add form validators.
   */
  addValidators(): void {
    this.formGroup.controls.id.addValidators([
      control => Validators.required(control),
      Validators.pattern(/^[a-zA-Z0-9]*$/),
    ]);
    this.formGroup.controls.id.updateValueAndValidity();
  }

  /**
   * Dispatch action to call job info service.
   */
  callJobInfoService(): void {
    this.awsService
      .callJobHistoryService<DynamoDBItem>(
        `${this.baseUrl}/${this.jobsService}`,
        this.formGroup.controls.id
      )
      .subscribe(jobHistoryResponse => {
        this.updateState({jobHistoryResponse});

        this.awsService.openTerminateDialog(jobHistoryResponse, () =>
          this.callService()
        );
      });
  }

  /**
   * Dispatch action to call terminate job service.
   */
  callService(): void {
    this.awsService
      .callTerminateJobService(
        `${this.baseUrl}/${this.terminateService}`,
        this.formGroup.controls.id
      )
      .subscribe(terminateJobResponse =>
        this.updateState({terminateJobResponse})
      );
  }

  /**
   * Return default control panel form values
   */
  defaultFormValues(): IdForm {
    return {
      id: null,
    };
  }

  /**
   * Dispatch action to initialize application.
   */
  init(): void {
    const query = this.route.snapshot.queryParams as IdForm;
    const formValues: IdForm = {
      ...query,
    };

    this.formGroup.patchValue(formValues);

    if (this.formGroup.valid) {
      this.nshmpService.selectPlotControl();
      this.callJobInfoService();
    }
  }

  /**
   * Returns initial application state.
   */
  initialState(): AppState {
    return {
      jobHistoryResponse: null,
      terminateJobResponse: null,
    };
  }

  /**
   * Update state.
   *
   * @param state Partial state to update
   */
  updateState(state: Partial<AppState>): void {
    this.state.set({
      ...this.state(),
      ...state,
    });
  }

  /**
   * Update URL based on form values.
   */
  private updateUrl(): void {
    const query: Partial<IdForm> = {
      id: this.formGroup.getRawValue().id,
    };

    this.router
      .navigate([devApps().aws.terminateHazJobs.routerLink], {
        queryParams: {
          ...query,
        },
      })
      .catch((error: Error) => this.nshmpService.throwError$(error));
  }
}
