import {AsyncPipe} from '@angular/common';
import {Component, computed, OnDestroy, OnInit} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatOption} from '@angular/material/core';
import {MatError, MatFormField, MatLabel} from '@angular/material/form-field';
import {MatInput} from '@angular/material/input';
import {MatSelect} from '@angular/material/select';
import {
  GmmFormControlIds,
  gmmUtils,
  MultiSelectableParam,
  MW_COMMON_VALUES,
  NshmpLibNgGmmMenuComponent,
  NshmpLibNgGmmMultiParamMenuComponent,
  NshmpLibNgGmmMultiSelectComponent,
  NshmpLibNgGmmPlotOptionsControlPanelComponent,
  VS30_COMMON_VALUES,
} from '@ghsc/nshmp-lib-ng/gmm';
import {
  NshmpLibNgControlPanelButtonsComponent,
  NshmpService,
} from '@ghsc/nshmp-lib-ng/nshmp';
import {combineLatest, Subscription} from 'rxjs';
import {map} from 'rxjs/operators';

import {AppService} from '../../services/app.service';

/**
 * Control panel form fields and buttons to call GMM vs distance service.
 */
@Component({
  imports: [
    MatFormField,
    MatLabel,
    MatSelect,
    MatOption,
    MatInput,
    MatError,
    AsyncPipe,
    NshmpLibNgGmmMenuComponent,
    NshmpLibNgGmmMultiSelectComponent,
    NshmpLibNgGmmPlotOptionsControlPanelComponent,
    NshmpLibNgControlPanelButtonsComponent,
    NshmpLibNgGmmMultiParamMenuComponent,
    ReactiveFormsModule,
  ],
  selector: 'app-control-panel',
  standalone: true,
  styleUrl: './control-panel.component.scss',
  templateUrl: './control-panel.component.html',
})
export class ControlPanelComponent implements OnInit, OnDestroy {
  imtPlaceHolder = gmmUtils.imtPlaceHolder();

  /** List of common magnitudes */
  MwCommonValues = MW_COMMON_VALUES;

  /** List of common vs30 values */
  vs30CommonValues = VS30_COMMON_VALUES;

  /** Form controls state */
  controls = this.service.controlPanelForm.controls;

  /** Control panel form field state */
  formState = this.service.controlPanelForm;

  /** Whether a GMM has been selected */
  gmmSelected$ = this.controls.gmmSource.valueChanges.pipe(
    map(gmmSource => gmmSource?.length > 0)
  );

  /** Usage parameters */
  parameters = computed(
    () => this.service.state().usageResponse?.response?.parameters
  );

  /** Service call info state */
  serviceCallInfo = computed(() => this.service.state().serviceCallInfo);

  /** Supported IMTs based on GMMs selected */
  supportedImts = computed(() => this.service.state().supportedImts);

  /** Whether magnitudes are multi-selectable */
  MwMultiple$ = this.controls.multiSelectableParam.valueChanges.pipe(
    map(
      multiSelectableParam =>
        multiSelectableParam.toString() === GmmFormControlIds.MW.toString()
    )
  );

  /** Whether Vs30 is multi-selectable */
  vs30Multiple$ = this.controls.multiSelectableParam.valueChanges.pipe(
    map(
      multiSelectableParam =>
        multiSelectableParam.toString() === GmmFormControlIds.VS30.toString()
    )
  );

  subs: Subscription[] = [];

  constructor(
    public service: AppService,
    private nshmpService: NshmpService
  ) {}

  ngOnInit(): void {
    this.subs.push(
      this.formState.controls.gmmSource.valueChanges.subscribe(() => {
        this.onGmmSource();
      })
    );

    this.subs.push(
      this.formState.controls.multiSelectableParam.valueChanges.subscribe(() =>
        this.onMultiSelectableParam()
      )
    );

    this.subs.push(
      this.formState.controls.showEpistemicUncertainty.valueChanges.subscribe(
        () => this.onShowEpistemicUncertainty()
      )
    );

    this.subs.push(
      combineLatest([
        this.formState.controls.dip.valueChanges,
        this.formState.controls.width.valueChanges,
      ]).subscribe(() => this.updatePlots())
    );

    this.subs.push(
      combineLatest([
        this.formState.controls.imt.valueChanges,
        this.formState.controls.Mw.valueChanges,
        this.formState.controls.MwMulti.valueChanges,
        this.formState.controls.vs30Multi.valueChanges,
        this.formState.controls.vs30.valueChanges,
        this.formState.controls.z1p0.valueChanges,
        this.formState.controls.z2p5.valueChanges,
        this.formState.controls.zSed.valueChanges,
      ]).subscribe(() => this.service.resetState())
    );
  }

  ngOnDestroy(): void {
    this.subs.forEach(sub => sub?.unsubscribe());
  }

  /**
   * On form submit.
   */
  onSubmit(): void {
    this.service.callService();
    this.nshmpService.selectPlotControl();
  }

  private onGmmSource() {
    const formValues = this.formState.getRawValue();

    if (
      !this.service.state().usageResponse ||
      formValues.gmmSource.length === 0
    ) {
      this.formState.controls.imt.patchValue(gmmUtils.imtPlaceHolder().value);
      this.service.updateState({
        supportedImts: [],
      });
      return;
    }

    const supportedImts = gmmUtils.getSupportedImts(
      formValues.gmmSource,
      this.service.state().usageResponse
    );

    this.service.updateState({
      supportedImts,
    });

    const imt = [...supportedImts].shift();

    this.formState.controls.imt.patchValue(imt?.value);
    this.service.resetState();
  }

  private onMultiSelectableParam(): void {
    if (!this.service.state().usageResponse) {
      return;
    }

    const {multiSelectableParam} = this.formState.getRawValue();
    const parameters = this.service.state().usageResponse.response.parameters;

    if (multiSelectableParam === MultiSelectableParam.MW) {
      this.formState.controls.Mw.setValue(parameters.Mw.value as number);
    } else if (multiSelectableParam === MultiSelectableParam.VS30) {
      this.formState.controls.vs30.setValue(parameters.vs30.value as number);
    }

    this.formState.controls.vs30Multi.setValue([]);
    this.formState.controls.MwMulti.setValue([]);
    this.formState.controls.gmmSource.setValue([]);
    this.service.resetState();
  }

  private onShowEpistemicUncertainty(): void {
    this.service.createPlots();
  }

  private updatePlots(): void {
    this.service.createPlots();
  }
}
