import {Injectable} from '@angular/core';
import {Router, UrlTree} from '@angular/router';
import {HazardService} from '@ghsc/nshmp-lib-ng/hazard';
import {NshmpService, SpinnerService} from '@ghsc/nshmp-lib-ng/nshmp';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {catchError, map, Observable, of} from 'rxjs';

/**
 * Guard to allow dynamic compare application launch.
 *
 * Will not be able to launch if there are no models to compare.
 */
@Injectable({
  providedIn: 'root',
})
export class DynamicHazardCompareGuard {
  /** Error message */
  errorMessage = `No comparable models available. <br>
    Application cannot be launch.`;
  /** nshmp-haz-ws web config */
  nshmpHazWs = environment.webServices.nshmpHazWs;
  /** Hazard endpoint */
  serviceEndpoint = this.nshmpHazWs.services.curveServices.hazard;

  constructor(
    private router: Router,
    private hazardService: HazardService,
    private spinnerService: SpinnerService,
    private nshmpService: NshmpService
  ) {}

  /**
   * Returns true if there are comparable models.
   */
  canActivate():
    | Observable<boolean | UrlTree>
    | Promise<boolean | UrlTree>
    | boolean
    | UrlTree {
    this.spinnerService.remove();
    this.spinnerService.show('Checking for comparable models');

    return this.hasComparableModels$.pipe(
      map(hasComparableModels => {
        this.spinnerService.remove();

        if (!hasComparableModels) {
          this.nshmpService.throwError$(new Error(this.errorMessage));

          this.router
            .navigate([''])
            .then(() => this.router.navigate(['dev']))
            .catch((error: Error) => this.nshmpService.throwError$(error));
        }

        return hasComparableModels;
      })
    );
  }

  /**
   * Check to see if there are any comparable models.
   */
  private get hasComparableModels$(): Observable<boolean> {
    return this.hazardService
      .dynamicNshms$(
        `${this.nshmpHazWs.url}${this.nshmpHazWs.services.nshms}`,
        this.serviceEndpoint
      )
      .pipe(
        map(dynamicNshms => {
          const hasComparableModels = dynamicNshms.nshmServices.some(
            nshmService =>
              dynamicNshms.nshmServices.find(
                service =>
                  service.project === nshmService.project &&
                  service.model !== nshmService.model
              )
          );

          return hasComparableModels;
        }),
        catchError(() => {
          this.spinnerService.remove();
          return of(false);
        })
      );
  }
}
