import {computed, Injectable, Signal, signal} from '@angular/core';
import {FormBuilder, FormControl, Validators} from '@angular/forms';
import {FormGroupControls} from '@ghsc/nshmp-lib-ng/nshmp';
import {
  NshmpPlot,
  NshmpPlotSettings,
  PlotOptions,
  plotUtils,
} from '@ghsc/nshmp-lib-ng/plot';
import {ExceedanceModel, Maths} from '@ghsc/nshmp-utils-ts/libs/nshmp-lib/calc';
import {UncertaintyModel} from '@ghsc/nshmp-utils-ts/libs/nshmp-lib/data';
import * as d3 from 'd3-array';
import deepEqual from 'deep-equal';
import {PlotData} from 'plotly.js';
import {AppServiceModel} from 'projects/nshmp-apps/src/shared/models/app-service.model';
import {SharedService} from 'projects/nshmp-apps/src/shared/services/shared.service';

import {ExceedanceFormBounds, FormBounds} from '../models/bounds.model';
import {ControlForm} from '../models/control-form.model';
import {Plots} from '../models/plots.model';
import {AppState} from '../models/state.model';

/**
 * Entrypoint for accessing store.
 */
@Injectable({
  providedIn: 'root',
})
export class AppService
  extends SharedService
  implements AppServiceModel<AppState, ControlForm>
{
  /** Control panels forms */
  readonly formGroup = this.formBuilder.group<ControlForm>(
    this.defaultFormValues()
  );

  readonly state = signal<AppState>(this.initialState());

  constructor(private formBuilder: FormBuilder) {
    super();
    this.addValidators();
  }

  /**
   * Returns the exceedance plot.
   */
  get plot(): Signal<NshmpPlot> {
    return computed(() => this.state().plots.get(Plots.EXCEEDANCE));
  }

  addValidators(): void {
    this.boundsValidators(
      this.formGroup.controls.median,
      this.formBounds().median
    );
    this.boundsValidators(this.formGroup.controls.rate, this.formBounds().rate);
    this.boundsValidators(
      this.formGroup.controls.sigma,
      this.formBounds().sigma
    );
    this.boundsValidators(
      this.formGroup.controls.truncationLevel,
      this.formBounds().truncationLevel
    );
    this.formGroup.controls.truncation.addValidators(control =>
      Validators.required(control)
    );
    this.formGroup.updateValueAndValidity();
  }

  /**
   * Calculate exceedance.
   */
  callService(): void {
    this.updateState({
      plots: this.addPlotData(this.state(), this.formGroup),
    });
  }

  /**
   * Clear the plot.
   */
  clearPlot(): void {
    this.updateState(this.initialState());
  }

  /**
   * Exceedance default control panel form values.
   */
  defaultFormValues(): ControlForm {
    return {
      median: 1.0,
      rate: 1,
      sigma: 0.5,
      truncation: true,
      truncationLevel: 3,
    };
  }

  /**
   * Exceedance X values for plot.
   */
  exceedanceXValues(): number[] {
    return d3.ticks(Math.log(0.0001), Math.log(10.0), 100);
  }

  /** Exceedance form bounds */
  formBounds(): ExceedanceFormBounds {
    return {
      median: {
        max: 10,
        min: 0,
        step: 0.5,
      },
      rate: {
        max: 1,
        min: 0,
        step: 0.1,
      },
      sigma: {
        max: 1,
        min: 0,
        step: 0.1,
      },
      truncationLevel: {
        max: 5,
        min: 0,
        step: 1,
      },
    };
  }

  init(): void {}

  /**
   * Initial exceedance app state.
   */
  initialState(): AppState {
    return {
      plots: this.defaultPlots(),
    };
  }

  /**
   * Reset the plot settings.
   */
  resetSettings(): void {
    super.resetPlotSettings({
      currentPlots: this.state().plots,
      defaultPlots: this.defaultPlots(),
    });
  }

  /**
   * Update a single plot in the state.
   *
   * @param plot Updated plot
   * @param id Id of plot to update
   */
  updatePlot(plot: NshmpPlot, id: string): void {
    const plots = new Map(this.state().plots);
    plots.set(id, plot);
    this.updateState({plots});
  }

  updateState(state: Partial<AppState>): void {
    const updatedState = {
      ...this.state(),
      ...state,
    };

    if (!deepEqual(updatedState, this.state())) {
      this.state.set({
        ...this.state(),
        ...state,
      });
    }
  }

  /**
   * Calculate exceedance and add plot data.
   *
   * @param state The current state
   */
  private addPlotData(
    state: AppState,
    formGroup: FormGroupControls<ControlForm>
  ): Map<string, NshmpPlot> {
    const exceedanceData = this.calculateExceedance(formGroup.getRawValue());
    const plots = new Map<string, NshmpPlot>();
    const plot = state.plots.get(Plots.EXCEEDANCE);
    const data = [...plot.plotData.data];
    data.push(exceedanceData);

    plots.set(Plots.EXCEEDANCE, {
      ...plot,
      plotData: {
        ...plot.plotData,
        data,
      },
    });

    return plots;
  }

  private boundsValidators(control: FormControl, bounds: FormBounds): void {
    control.addValidators([
      Validators.min(bounds.min),
      Validators.max(bounds.max),
      c => Validators.required(c),
    ]);
  }

  /**
   * Calculate exceedance and return the plot data.
   *
   * @param values The control form values
   * @returns
   */
  private calculateExceedance(values: ControlForm): Partial<PlotData> {
    const model: UncertaintyModel = {
      n: values.truncationLevel,
      μ: Math.log(values.median),
      σ: values.sigma,
    };

    const label = `μ=${values.median}, σ=${values.sigma}, rate=${values.rate}`;

    const yValues: number[] = values.truncation
      ? ExceedanceModel.truncationUpperOnlySequence(
          model,
          this.exceedanceXValues()
        )
      : ExceedanceModel.truncationOffSequence(model, this.exceedanceXValues());

    return {
      mode: 'lines+markers',
      name: values.truncation ? `${label}, n=${values.truncationLevel}` : label,
      x: this.exceedanceXValues().map(x => Math.exp(x)),
      y: yValues.map(y => Maths.round(y * values.rate, 4)),
    };
  }

  private defaultPlots(): Map<string, NshmpPlot> {
    const plots = new Map<string, NshmpPlot>();

    const plotOptions: PlotOptions = {
      layout: {
        legend: {
          bordercolor: ' #a9aeb1',
          borderwidth: 1,
          x: 0.01,
          y: 0.1,
          yanchor: 'bottom',
        },
        yaxis: {
          range: [0, 1],
          tickmode: 'auto',
          type: 'linear',
        },
      },
    };

    const plotData = plotUtils.defaultPlot({
      id: 'exceedance-explorer',
      mobileOptions: {
        ...plotOptions,
        layout: {
          ...plotOptions.layout,
          margin: {
            b: 50,
          },
        },
      },
      options: {...plotOptions},
      title: 'Exceedance',
      xLabel: 'Ground Motion (g)',
      yLabel: 'Annual Frequency of Exceedence',
    });

    const settingsForm: NshmpPlotSettings = {
      config: plotData.config,
      layout: plotUtils.plotlyLayoutToSettings(plotData.layout),
    };

    plots.set(Plots.EXCEEDANCE, {
      label: 'Exceedance',
      plotData,
      settingsForm: plotUtils.plotSettingsToFormGroup(settingsForm),
    });

    return plots;
  }
}
