import {AsyncPipe} from '@angular/common';
import {Component, computed, OnDestroy, OnInit, Signal} from '@angular/core';
import {ReactiveFormsModule} from '@angular/forms';
import {MatCheckbox} from '@angular/material/checkbox';
import {MatOption} from '@angular/material/core';
import {MatError, MatFormField, MatLabel} from '@angular/material/form-field';
import {MatInput} from '@angular/material/input';
import {MatRadioButton, MatRadioGroup} from '@angular/material/radio';
import {MatSelect} from '@angular/material/select';
import {
  DisaggTarget,
  NshmpLibNgHazardLocationFormComponent,
  NshmpLibNgHazardModelFormComponent,
  NshmpLibNgHazardReturnPeriodFormComponent,
  NshmpLibNgHazardSiteClassFormComponent,
} from '@ghsc/nshmp-lib-ng/hazard';
import {
  NshmpLibNgMapSelectSiteComponent,
  SelectSiteDialogData,
} from '@ghsc/nshmp-lib-ng/map';
import {
  NshmpLibNgControlPanelButtonsComponent,
  NshmpService,
} from '@ghsc/nshmp-lib-ng/nshmp';
import {environment} from 'projects/nshmp-apps/src/environments/environment';
import {combineLatest, Subscription} from 'rxjs';

import {AppService} from '../../services/app.service';

/**
 * Control panel with form fields to call disagg service.
 *
 * Handles disagg service with return period and imt.
 */
@Component({
  imports: [
    NshmpLibNgHazardModelFormComponent,
    NshmpLibNgHazardLocationFormComponent,
    NshmpLibNgMapSelectSiteComponent,
    NshmpLibNgHazardSiteClassFormComponent,
    NshmpLibNgControlPanelButtonsComponent,
    NshmpLibNgHazardReturnPeriodFormComponent,
    NshmpLibNgControlPanelButtonsComponent,
    MatFormField,
    MatLabel,
    MatSelect,
    MatOption,
    MatRadioGroup,
    MatRadioButton,
    MatInput,
    MatError,
    MatCheckbox,
    AsyncPipe,
    ReactiveFormsModule,
  ],
  selector: 'app-control-panel',
  standalone: true,
  styleUrl: './control-panel.component.scss',
  templateUrl: './control-panel.component.html',
})
export class ControlPanelComponent implements OnInit, OnDestroy {
  /** Disaggregation target enum */
  DisaggTarget = DisaggTarget;

  /** Form fields state */
  form = this.service.formGroup;

  /** IML range from usage, based on IMT value */
  imlRange = computed(() => {
    const usage = this.service.usage();

    if (usage) {
      return usage.response.iml.range[this.form.getRawValue().imt];
    }
  });

  /** Data for select site component */
  selectSiteData: Signal<SelectSiteDialogData> = computed(() => {
    const usage = this.service.usage();
    const nshmService = this.service.nshmService();

    if (usage === null) {
      return;
    }

    const services = environment.webServices.nshmpHazWs.services.curveServices;

    return {
      mapUrl: `${nshmService.url}${services.map}?raw=true`,
      nshm: this.form.getRawValue().model,
      sitesUrl: `${nshmService.url}${services.sites}?raw=true`,
    };
  });

  /** Service call info state */
  serviceCallInfo = this.service.serviceCallInfo;

  /** The site classes */
  siteClasses = computed(
    () => this.service.usage()?.response.model.siteClasses
  );

  /** Source model info from usage */
  sourceModel = computed(() => this.service.usage()?.response?.model);

  private subs: Subscription[] = [];

  constructor(
    public service: AppService,
    private nshmpService: NshmpService
  ) {}

  ngOnInit(): void {
    const controls = this.form.controls;

    this.subs.push(
      controls.disaggTarget.valueChanges.subscribe(() => {
        this.service.onDisaggTarget();
        this.service.resetState();
      })
    );

    this.subs.push(
      controls.model.valueChanges.subscribe(() => this.onModelChange())
    );

    this.subs.push(
      combineLatest([
        controls.iml.valueChanges,
        controls.imt.valueChanges,
        controls.latitude.valueChanges,
        controls.longitude.valueChanges,
        controls.returnPeriod.valueChanges,
        controls.vs30.valueChanges,
      ]).subscribe(() => {
        this.service.resetState();
      })
    );
  }

  ngOnDestroy(): void {
    this.subs.forEach(sub => sub?.unsubscribe());
  }

  private onModelChange(): void {
    const {latitude, longitude} = this.form.getRawValue();
    const latitudeBounds = this.service.usage().response.latitude;
    const longitudeBounds = this.service.usage().response.longitude;

    if (
      !(
        latitude >= latitudeBounds.min &&
        latitude <= latitudeBounds.max &&
        longitude >= longitudeBounds.min &&
        longitude <= longitudeBounds.max
      )
    ) {
      this.form.patchValue({
        latitude: NaN,
        longitude: NaN,
      });
    }

    this.form.controls.latitude.markAsPristine();
    this.form.controls.longitude.markAsPristine();
    this.service.resetState();
  }

  /**
   * On form submit.
   */
  onSubmit(): void {
    this.service.callService();
    this.nshmpService.selectPlotControl();
  }
}
