import {AsyncPipe} from '@angular/common';
import {Component, computed} from '@angular/core';
import {MatDivider} from '@angular/material/divider';
import {
  MatAccordion,
  MatExpansionPanel,
  MatExpansionPanelContent,
  MatExpansionPanelHeader,
} from '@angular/material/expansion';
import {HazardControlForm, hazardUtils} from '@ghsc/nshmp-lib-ng/hazard';
import {
  NshmpLibNgDataTableComponent,
  NshmpLibNgExportDataTableComponent,
  NshmpService,
  TableData,
} from '@ghsc/nshmp-lib-ng/nshmp';
import {Imt} from '@ghsc/nshmp-utils-ts/libs/nshmp-lib/gmm';
import {
  HazardResponseMetadata,
  StaticResponseData,
} from '@ghsc/nshmp-utils-ts/libs/nshmp-ws-static/hazard-service';

import {StaticHazardCurveTable} from '../../models/curve-table.model';
import {AppService} from '../../services/app.service';

/**
 * Tables of static hazard data.
 */
@Component({
  imports: [
    NshmpLibNgExportDataTableComponent,
    NshmpLibNgDataTableComponent,
    MatAccordion,
    MatExpansionPanel,
    MatExpansionPanelHeader,
    MatDivider,
    MatExpansionPanelContent,
    AsyncPipe,
  ],
  selector: 'app-curve-data',
  standalone: true,
  styleUrl: './curve-data.component.scss',
  templateUrl: './curve-data.component.html',
})
export class CurveDataComponent {
  /** The data precision */
  dataPrecision = 4;

  /** Static hazard service reponse */
  serviceResponse = this.service.serviceResponse;

  formGroup = this.service.formGroup;

  /** Static hazard curve tables */
  tables = computed(() => {
    if (this.serviceResponse() === null) {
      return [];
    }
    this.service.plots();
    return this.responseDataToTable(
      this.serviceResponse().response,
      this.service.formGroup.getRawValue()
    );
  });

  /** Static hazard table data  */
  tableData = computed(() => this.tablesToTableData(this.tables()));

  constructor(
    private service: AppService,
    public nshmpService: NshmpService
  ) {}

  /** Filename for hazard export CSV */
  hazardFilename(): string {
    const {latitude, longitude, model} = this.service.formGroup.getRawValue();

    return `static-hazard-curves-${model}-(${longitude},${latitude}).csv`;
  }

  /**
   * Convert response data to curve table data.
   *
   * @param responseData The response data
   * @param form Form values
   * @returns
   */
  private responseDataToTable(
    responseData: StaticResponseData<HazardResponseMetadata>[][],
    form: HazardControlForm
  ): StaticHazardCurveTable[] {
    return responseData.map(response => {
      const tableData = this.responseDataToTableData(response, form);
      const siteClass = [...response].pop().metadata.siteClass;
      return {
        siteClass,
        tableData,
      };
    });
  }

  /**
   * Convert response data to table table.
   *
   * @param responseData The response data
   * @param form Form values
   */
  private responseDataToTableData(
    responseData: StaticResponseData<HazardResponseMetadata>[],
    form: HazardControlForm
  ): TableData[] {
    const tableData: TableData[] = [];
    responseData.forEach((response, index) => {
      const imt = response.metadata.imt.value as Imt;
      const xy = hazardUtils.updateXySequence(form, response.data, imt);

      const xData: TableData = {
        label: response.metadata.imt.display,
        td: xy.xs.map(x => (x ? x.toFixed(this.dataPrecision) : x)),
        th: response.metadata.xLabel,
      };
      if (index === 0) {
        xData.exportSectionBreakLabel = `Site class - ${response.metadata.siteClass}`;
      }
      tableData.push(xData);

      tableData.push({
        td: xy.ys.map(y => (y ? y.toExponential(this.dataPrecision) : y)),
        th: response.metadata.yLabel,
      });
    });

    return tableData;
  }

  /**
   * Convert static curve tables to table data.
   *
   * @param tables The hazard curve table
   */
  private tablesToTableData(tables: StaticHazardCurveTable[]): TableData[] {
    if (tables === null || tables.length === 0) {
      return [];
    }
    return tables
      .map(table => table.tableData)
      .reduce((previouse, current) => [...previouse, ...current]);
  }
}
