import {AsyncPipe} from '@angular/common';
import {Component, computed} from '@angular/core';
import {
  MatAccordion,
  MatExpansionPanel,
  MatExpansionPanelHeader,
} from '@angular/material/expansion';
import {hazardUtils, ResponseSpectra} from '@ghsc/nshmp-lib-ng/hazard';
import {
  NshmpLibNgDataTableComponent,
  NshmpLibNgExportDataTableComponent,
  TableData,
} from '@ghsc/nshmp-lib-ng/nshmp';

import {StaticHazardCurveTable} from '../../models/curve-table.model';
import {AppService} from '../../services/app.service';

/**
 * Tables spectra data.
 */
@Component({
  imports: [
    NshmpLibNgExportDataTableComponent,
    NshmpLibNgDataTableComponent,
    MatAccordion,
    MatExpansionPanel,
    MatExpansionPanelHeader,
    AsyncPipe,
  ],
  selector: 'app-spectrum-data',
  standalone: true,
  styleUrl: './spectrum-data.component.scss',
  templateUrl: './spectrum-data.component.html',
})
export class SpectrumDataComponent {
  /** The response spectra data */
  responseSpectra = this.service.responseSpectra;

  formGroup = this.service.formGroup;

  /** Static hazard curve table */
  tables = computed(() => {
    if (this.responseSpectra()) {
      return this.spectraToTableData(this.responseSpectra());
    } else {
      return [];
    }
  });

  /** Static hazard table data */
  tableData = computed(() => this.tablesToTableData(this.tables()));

  constructor(private service: AppService) {}

  /** Spectra export filename */
  spectraFilename(): string {
    const {latitude, longitude, model} = this.service.formGroup.getRawValue();

    return `static-hazard-spectra-${model}-(${longitude},${latitude}).csv`;
  }

  /**
   * Convert response spectra data to table data.
   *
   * @param spectra Response spectra data
   */
  private spectraToTableData(
    spectra: ResponseSpectra[]
  ): StaticHazardCurveTable[] {
    return spectra.map(spectrum => {
      const tableData = hazardUtils
        .responseSpectraToTableData(spectrum)
        .map((data, index) => {
          if (index === 0) {
            data.exportSectionBreakLabel = `Site Class - ${spectrum.siteClass}`;
          }
          return data;
        });

      return {
        siteClass: spectrum.siteClass,
        tableData,
      };
    });
  }

  /**
   * Convert static hazard curve tables to table data.
   *
   * @param tables Curve tables
   */
  private tablesToTableData(tables: StaticHazardCurveTable[]): TableData[] {
    if (tables === null || tables.length === 0) {
      return [];
    }
    return tables
      .map(table => table.tableData)
      .reduce((previouse, current) => [...previouse, ...current]);
  }
}
